import json


JSON_NULL = object()


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if obj is JSON_NULL:
            return None
        return super(JSONEncoder, self).default(obj)



import logging

from collections import namedtuple

from django.apps import apps
from django.db import connection

from wicked_historian.models import BaseEditHistory


SUPPORTED_DATABASES = ('postgresql', 'mysql')


logger = logging.getLogger(__name__)


ViewRow = namedtuple('ViewRow', 'db_table field_id field_name')


def _prepare_view_rows_from_models():  # type: () -> list[ViewRow]
    rows = []

    history_models = [model for model in apps.get_models() if issubclass(model, BaseEditHistory)]
    for model in history_models:
        for field_description in model.FIELDS_DESCRIPTIONS:
            attrs = {
                'db_table': model._meta.db_table,
                'field_id': field_description.id,
                'field_name': field_description.name
            }
            rows.append(ViewRow(**attrs))

    return rows


def _prepare_sql_view_code_from_rows(rows):  # type: (list[ViewRow]) -> (str, list)
    """Compose SQL code along with values to create or update a view for specified rows."""
    fields = ViewRow._fields

    sql_header = 'CREATE OR REPLACE VIEW wicked_historian_managed_fields ({}) AS '.format(', '.join(fields))
    sql_row = 'SELECT ' + ', '.join(['%s'] * len(fields))
    sql = sql_header + ' UNION '.join([sql_row] * len(rows))
    values = [getattr(row, field) for row in rows for field in fields]

    return sql, values


def create_choices_db_view(sender, **kwargs):
    """Create one db view for all wicked historian serviced django models."""
    if connection.vendor not in SUPPORTED_DATABASES:
        logger.warning('Your database is not supported, the view will not be created')
        return

    view_rows = _prepare_view_rows_from_models()
    if not view_rows:
        return

    sql, values = _prepare_sql_view_code_from_rows(view_rows)

    with connection.cursor() as cursor:
        cursor.execute(sql, values)



import threading
from contextlib import contextmanager
from typing import (
    Callable,
    Optional,
    Union,
)

from django.contrib.auth.base_user import AbstractBaseUser
from django.http import (
    HttpRequest,
    HttpResponse,
)

__all__ = (
    'usersmuggler',
    'NoUserSetException',
    'UserSmugglerMiddleware',
)


class NoUserSetException(Exception):
    pass


class UserSmuggler(object):

    class NotSetType(object):

        """A type for the NOT_SET object."""

    NOT_SET = NotSetType()
    _storage = threading.local()

    @contextmanager
    def set_user(self, user: Optional[AbstractBaseUser]):
        previous_user = self._get_stored_user()
        self._set_stored_user(user)
        try:
            yield
        finally:
            self._set_stored_user(previous_user)

    def get_user(self) -> Optional[AbstractBaseUser]:
        stored_user = self._get_stored_user()

        if stored_user is self.NOT_SET:
            raise NoUserSetException(
                "User was not specified, please use `UserSmuggler.set_user()` context outside of call to `get_user()`.")

        return stored_user

    def _get_stored_user(self) -> Union[AbstractBaseUser, None, NotSetType]:
        try:
            return self._storage.user
        except AttributeError:
            return self.NOT_SET

    def _set_stored_user(self, user: Union[AbstractBaseUser, None, NotSetType]):
        self._storage.user = user


usersmuggler = UserSmuggler()


class UserSmugglerMiddleware:

    """A middleware setting usersmuggler user."""

    def __init__(self, get_response: Callable):
        """Store the get_response callback in the middleware instance."""
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """Set current user in usersmuggler."""
        if request.user.is_authenticated:
            with usersmuggler.set_user(request.user):
                return self.get_response(request)
        return self.get_response(request)



import threading
from collections import namedtuple
from contextlib import contextmanager
from typing import (
    Any,
    List,
)


SignalExclusionKey = namedtuple('SignalExclusionKey', ['model_class', 'model_id', 'field_name'])


class SignalExclusion:

    """Object used for signals exclusion."""

    _storage = threading.local()

    @contextmanager
    def model_signals_exclusion_context(self, model_class: Any, model_id: Any, field_name: str):
        """Context in which for supplied "identification" of instance-field pair model signals are marked to be excluded from history handling."""
        try:
            self.store_exclusion_key(model_class, model_id, field_name, 'model_signals')
            yield
        finally:
            self.remove_exclusion_key(model_class, model_id, field_name, 'model_signals')

    def are_model_signals_excluded(self, model_class: Any, model_id: Any, field_name: str) -> bool:
        """Check if model signals for supplied args are excluded."""
        return self.is_exclusion_key_present(model_class, model_id, field_name, 'model_signals')

    def is_exclusion_key_present(self, model_class: Any, model_id: Any, field_name: str, signal_type: str) -> bool:
        """Check if key formed from supplied args is present in store for appropriate signal type."""
        store = self._get_store(signal_type)
        return SignalExclusionKey(model_class, model_id, field_name) in store

    def store_exclusion_key(self, model_class: Any, model_id: Any, field_name: str, signal_type: str):
        """Store key formed from supplied args in store for appropriate signal type."""
        store = self._get_store(signal_type)
        store.insert(0, SignalExclusionKey(model_class, model_id, field_name))

    def remove_exclusion_key(self, model_class: Any, model_id: Any, field_name: str, signal_type: str):
        """Remove key formed from supplied args from store for appropriate signal type."""
        try:
            store = getattr(self._storage, signal_type)
        except AttributeError:
            return
        else:
            store.remove(SignalExclusionKey(model_class, model_id, field_name))

    def _get_store(self, signal_type: str) -> List[SignalExclusionKey]:
        """Get or create store for appropriate signal type."""
        try:
            store = getattr(self._storage, signal_type)
        except AttributeError:
            store = []
            setattr(self._storage, signal_type, store)

        return store


signal_exclusion = SignalExclusion()



import json

from django.contrib.postgres.fields import JSONField as PostgresJSONField

from wicked_historian.encoder import JSONEncoder


__all__ = (
    'JSONField',
)


class JSONField(PostgresJSONField):

    def __init__(self, *args, **kwargs):
        kwargs['encoder'] = kwargs.get('encoder', JSONEncoder)
        super().__init__(*args, **kwargs)




# -*- coding: utf-8 -*-

from django.contrib.auth.models import User
from django.test.testcases import TestCase

from wicked_historian.usersmuggler import (
    NoUserSetException,
    usersmuggler,
)


class UserSmugglerTestCase(TestCase):

    def test_no_user_has_been_specified(self):
        with self.assertRaises(NoUserSetException):
            usersmuggler.get_user()

    def test_none_has_been_specified_instead_of_user(self):
        with usersmuggler.set_user(user=None):
            self.assertIsNone(usersmuggler.get_user())

    def test_user_has_been_specified(self):
        test_user = User()

        with usersmuggler.set_user(user=test_user):
            self.assertEqual(usersmuggler.get_user(), test_user)

    def test_several_users_have_been_specified(self):
        test_user1 = User()
        test_user2 = User()

        with usersmuggler.set_user(user=test_user1):
            with usersmuggler.set_user(user=None):
                with usersmuggler.set_user(user=test_user2):
                    self.assertEqual(usersmuggler.get_user(), test_user2)
                self.assertIsNone(usersmuggler.get_user())
            self.assertEqual(usersmuggler.get_user(), test_user1)



import json

from django_mysql.models.fields import JSONField as MySQLJSONField
from wicked_historian.encoder import JSONEncoder


__all__ = (
    'JSONField',
)


class JSONField(MySQLJSONField):

    def __init__(self, *args, **kwargs):
        self.encoder = kwargs.pop('encoder', JSONEncoder)
        super().__init__(*args, **kwargs)

    def get_prep_value(self, value):
        if value is not None and not isinstance(value, str):
            return json.dumps(value, cls=self.encoder, allow_nan=False)

        return value

    def get_db_prep_value(self, value, connection, prepared=False):
        if not prepared and value is not None:
            return json.dumps(value, cls=self.encoder, allow_nan=False)
        return value



import base64
import hashlib
from datetime import (
    date,
    time,
    timedelta,
)
from decimal import Decimal
from functools import (
    singledispatch,
    update_wrapper,
    wraps,
)
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import (  # pylint: disable=unused-import
    Field,
    ManyToOneRel,
    Model,
)
from django.db.models.fields.files import FieldFile
from django.db.models.signals import (
    m2m_changed,
    pre_delete,
    pre_save,
    post_delete,
    post_save,
)
from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.module_loading import import_string

from .encoder import JSON_NULL
from .models import (
    BaseEditHistory,
    DiffableHistoryModel,
)
from .signals_exclusion import signal_exclusion
from .usersmuggler import usersmuggler


JSON_FIELD_KWARGS = getattr(settings, 'WICKED_HISTORIAN_JSON_FIELD_KWARGS', {})
json_field_class = import_string(settings.WICKED_HISTORIAN_JSON_FIELD_CLASS)  # type: Type[Field]
FieldValue = TypeVar('FieldValue')
ModelInstanceRepresentation = Dict[str, Any]
ValueRepresentation = Union[FieldValue, ModelInstanceRepresentation, Iterable[ModelInstanceRepresentation]]
ModelDiffItem = Tuple[str, Tuple[Any, Any]]
BaseModel = TypeVar('BaseModel')


class ReverseForeignKeyRelation(Field):
    """
    Class to mark reverse foreign key relations in field value mapper.
    """
    pass


class FieldDescription:
    __slots__ = ['name', 'field_instance']

    def __init__(self, name: str, field_instance: models.Field):
        self.name = name
        if not isinstance(field_instance, models.Field):
            raise ValueError("Field instance must be instance of django's field")
        self.field_instance = field_instance

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        return self.get_field_description_id(self.name, self.field_instance.__class__)

    @property
    def verbose_name(self) -> str:
        return self.field_instance.verbose_name or self.name

    @staticmethod
    def get_field_description_id(field_name: str, field_class: Type[models.Field]) -> str:
        class_path = field_class.__module__ + "." + field_class.__name__
        field_id = hashlib.sha1('name:{},field_instance:{}'.format(field_name, class_path).encode()).hexdigest()
        return field_id


class ObsoleteFieldDescription(FieldDescription):

    def __init__(self, name: str, field_instance: models.Field):
        super().__init__(name, field_instance)
        self.field_instance.set_attributes_from_name(self.name)


def generate_history_class(
        model_class: Type[DiffableHistoryModel],
        module: str,
        excluded_fields: Optional[Iterable[str]] = None,
        obsolete_field_choices: Optional[List[ObsoleteFieldDescription]] = None,
        abstract: bool = False,
) -> Type[BaseEditHistory]:

    def get_reverse_fk_relations_for_choices():
        return [i for i in getattr(model_class._meta, 'reverse_foreign_key_relations', set())]

    if not issubclass(model_class, DiffableHistoryModel):
        raise ImproperlyConfigured('%s class should inherit from DiffableHistoryModel' % model_class.__name__)
    excluded_fields = excluded_fields or []
    obsolete_field_choices = obsolete_field_choices or []

    fields_from_model = get_field_choices_from_model(model_class)
    reverse_m2m_field_choices = [FieldDescription(x, ReverseForeignKeyRelation()) for x in get_reverse_fk_relations_for_choices()]
    field_choices = fields_from_model + reverse_m2m_field_choices + obsolete_field_choices

    assert len(field_choices) == len(set([f.id for f in field_choices])), 'id (name and type) should be either in model or in obsolete and unique'

    field_choices_by_id = {description.id: description for description in field_choices}
    tracked_fields_choices_by_name = {
        description.name: description
        for description in fields_from_model + reverse_m2m_field_choices
        if description.name not in excluded_fields
    }

    class EditHistory(BaseEditHistory):
        """History of (diffable) model edits"""

        class Meta:
            abstract = True

        model = models.ForeignKey(model_class, on_delete=models.CASCADE)
        user = models.ForeignKey(
            settings.AUTH_USER_MODEL,
            blank=True,
            null=True,
            on_delete=models.PROTECT,
        )
        change_date = models.DateTimeField(default=timezone.now)
        field = models.CharField(
            max_length=40,
            choices=[(description.id, description.verbose_name) for description in field_choices],
        )
        old_value = json_field_class(**JSON_FIELD_KWARGS)
        new_value = json_field_class(**JSON_FIELD_KWARGS)

        FIELD_VALUE_MAPPER = DefaultFieldValueMapper()

        FIELDS_DESCRIPTIONS = field_choices  # Here we want all fields from model (even those excluded) and obsolete
        # Because history may contain entries from field, which become exclueded.

        @classmethod
        def recalculate_choices(cls):
            field_choices = obsolete_field_choices + get_field_choices_from_model(model_class) + reverse_m2m_field_choices
            cls._meta.get_field('field').choices = [(description.id, description.verbose_name) for description in field_choices]

        @classmethod
        def get_for(cls, instance: model_class) -> Iterable[Dict[str, Any]]:
            history_qs = cls.objects.select_related('user').filter(model=instance).order_by('-id')
            history = []
            for entry in history_qs:
                history.append(cls.get_history_entry(entry))
            return history

        @classmethod
        def get_history_entry(cls, entry: 'EditHistory') -> Dict[str, Any]:
            # noinspection PyProtectedMember
            try:
                field = field_choices_by_id[entry.field].field_instance
            except KeyError:
                raise cls.UnknownFieldException('Field \'{}\' is neither in model nor in obsolete fields'.format(entry.field))

            old_value = entry.old_value
            new_value = entry.new_value

            if field.flatchoices:
                old_value = force_text(dict(field.flatchoices).get(old_value, old_value), strings_only=True)
                new_value = force_text(dict(field.flatchoices).get(new_value, new_value), strings_only=True)

            # noinspection PyProtectedMember
            return {
                'user': entry.user,
                'change_date': entry.change_date,
                'field_verbose_name': entry.field_verbose_name,
                'old_value': old_value,
                'new_value': new_value,
            }

        @classmethod
        def get_tracked_field_choice_by_name(cls, name: str) -> FieldDescription:
            try:
                return tracked_fields_choices_by_name[name]
            except KeyError:
                raise cls.FieldNotTracked('Field \'{}\' is not known or is excluded or is obsolete'.format(name))

        @classmethod
        def get_value_representation(cls, name: str, value: FieldValue) -> ValueRepresentation:
            field = cls.get_tracked_field_choice_by_name(name).field_instance
            return cls.FIELD_VALUE_MAPPER(field, value)

        @classmethod
        def create_history(cls, instance: model_class, diff_items: List[ModelDiffItem]):
            # noinspection PyProtectedMember
            # pylint: disable=protected-access
            user = usersmuggler.get_user()
            for field_name, values in diff_items:
                try:
                    old_value = cls.get_value_representation(field_name, values[0])
                    new_value = cls.get_value_representation(field_name, values[1])
                    cls.objects.create(
                        model=instance,
                        field=cls.get_tracked_field_choice_by_name(field_name).id,
                        old_value=old_value,
                        new_value=new_value,
                        user=user,
                    )
                except cls.FieldNotTracked:
                    pass

            # ensure this properties are empty for further m2m checks
            instance._wicked_historian_last_m2m_record = None
            instance._wicked_historian_m2m_changes = {}

        @classmethod
        def create_m2m_history(cls, instance: model_class):
            if model_class.deletion_guard.is_instance_during_deletion(instance):
                return  # Don't trace changes on m2m during instance deletion

            user = usersmuggler.get_user()
            # noinspection PyProtectedMember
            # pylint: disable=protected-access
            m2m_changes = instance._wicked_historian_m2m_changes
            old_fields = m2m_changes['old_value']
            new_fields = m2m_changes['new_value']

            last_m2m_record = getattr(instance, '_wicked_historian_last_m2m_record', None)

            if last_m2m_record and last_m2m_record.field_name == m2m_changes['field_name']:
                last_m2m_record.new_value = new_fields
                last_m2m_record.save(update_fields=['new_value'])
            else:
                try:
                    instance._wicked_historian_last_m2m_record = cls.objects.create(
                        model=instance,
                        field=cls.get_tracked_field_choice_by_name(m2m_changes['field_name']).id,
                        old_value=old_fields,
                        new_value=new_fields,
                        user=user,
                    )
                except cls.FieldNotTracked:
                    pass

        @classmethod
        def m2m_changed(cls, sender: Type[Model], **kwargs):
            # pylint: disable=protected-access

            action = kwargs['action']
            instance = kwargs['instance']

            if '_' in sender.__name__:
                field_name = sender.__name__.split('_', 1)[1]
            else:
                # We gonna retrieve field name the hard way
                model = kwargs['model']
                # noinspection PyProtectedMember
                field_name = [
                    field.name for field in instance._meta.many_to_many if field.remote_field.model == model and field.remote_field.through == sender
                ][0]

            value = list(getattr(instance, field_name).all())

            if action in ['pre_add', 'pre_remove', 'pre_clear']:
                instance._wicked_historian_m2m_changes = {
                    'field_name': field_name,
                    'old_value': cls.get_value_representation(field_name, value)
                }
            elif action in ['post_add', 'post_remove', 'post_clear']:
                instance._wicked_historian_m2m_changes['new_value'] = cls.get_value_representation(field_name, value)
                cls.create_m2m_history(instance)

        @classmethod
        def get_foreign_key_field_name(cls, model_class_for_access_field_name: Type[Model], target_model: Type[Model]) -> Optional[str]:
            for field in model_class_for_access_field_name._meta.get_fields():  # pylint: disable=protected-access
                if (isinstance(field, models.ForeignKey) and
                    (
                            hasattr(field, 'rel') and field.rel.to == target_model or
                            hasattr(field, 'remote_field') and field.remote_field.model == target_model
                    )
                ):
                    return field.name
            return None

        @classmethod
        def get_model_or_none(cls, instance: Model) -> Optional[Model]:
            try:
                return instance.__class__.objects.get(pk=instance.pk)
            except instance.__class__.DoesNotExist:
                return None

        @classmethod
        def custom_through_m2m_handlers_factory(cls, model: Type[Model], m2m_field: Type[Field]) -> Tuple[Callable, Callable]:
            def get_fk_name_to_tracked_model(instance: Model) -> str:
                """Class to retrieve field name. It handles case, when there is more than one FK to same Model in custom through.

                :param instance: custom m2m instance
                :type instance: Model
                :return: field name of custom m2m which refers to current model.
                :rtype: str
                """
                through_field = getattr(model_class, m2m_field.name).rel.through_fields
                if through_field is not None:
                    return through_field[0]
                else:
                    return cls.get_foreign_key_field_name(instance.__class__, model)

            def get_instance_with_tracked_history(through_instance: Model, field_name: str, fresh_through_instance: Model) -> Model:
                instance_with_tracked_history = getattr(fresh_through_instance, field_name)
                # we retrieve an instance from the protected attribute or from the database if there is no data stored for the field
                instance_from_cache = through_instance._wicked_historian_changes_for_related_m2m_models.get(field_name, {}).get(
                    instance_with_tracked_history.pk, {}).get('instance')
                instance_with_tracked_history = instance_from_cache or instance_with_tracked_history
                return instance_with_tracked_history

            def get_m2m_value(instance_with_tracked_history: Model) -> ValueRepresentation:
                related_through_instances = list(getattr(instance_with_tracked_history, m2m_field.name).all())
                return cls.get_value_representation(m2m_field.name, related_through_instances)

            def are_signals_excluded(instance: Model) -> bool:
                """Check if inside model signals exclusion context of this instance and field.

                Used to exclude model signals from handling history creation.
                """
                field_name = get_fk_name_to_tracked_model(instance)
                instance._wicked_historian_changes_for_related_m2m_models = getattr(
                    instance, '_wicked_historian_changes_for_related_m2m_models', {},
                )
                instance_with_tracked_history = get_instance_with_tracked_history(instance, field_name, instance)
                return signal_exclusion.are_model_signals_excluded(model_class, instance_with_tracked_history.pk, m2m_field.name)

            def fill_wicked_historian_changes_for_related_m2m_models(through_instance: Model, fresh_through_instance: Optional[Model]=None):
                # fresh_through_instance is other than through_instance only
                # when custom m2m was pinned from through_instance to fresh_through_instance
                fresh_through_instance = fresh_through_instance or through_instance
                field_name = get_fk_name_to_tracked_model(through_instance)
                through_instance._wicked_historian_changes_for_related_m2m_models = getattr(
                    through_instance, '_wicked_historian_changes_for_related_m2m_models', {},
                )

                instance_with_tracked_history = get_instance_with_tracked_history(through_instance, field_name, fresh_through_instance)
                previous_value = get_m2m_value(instance_with_tracked_history)
                if getattr(instance_with_tracked_history, '_wicked_historian_m2m_changes', {}).get('field_name') != m2m_field.name:
                    instance_with_tracked_history._wicked_historian_m2m_changes = {
                        'field_name': m2m_field.name,
                        'old_value': previous_value,
                    }

                log = through_instance._wicked_historian_changes_for_related_m2m_models
                log.setdefault(field_name, {})
                log[field_name][instance_with_tracked_history.pk] = {
                    'previous_value': previous_value,  # previous value is used to avoid creating transitions to same state
                    'instance': instance_with_tracked_history,
                }

            def pre_action(sender: Type[Model], instance: Model, **kwargs):  # pylint: disable=unused-argument
                if are_signals_excluded(instance):
                    return
                instance_from_db = cls.get_model_or_none(instance)
                # we have to save changes for old and new instance if m2m was pinned from old to new instance
                # if m2m wasn't pinned then operation is idempotent
                fill_wicked_historian_changes_for_related_m2m_models(instance)
                fill_wicked_historian_changes_for_related_m2m_models(instance, instance_from_db)

            def post_action(sender: Type[Model], instance: Model, **kwargs):  # pylint: disable=unused-argument
                if are_signals_excluded(instance):
                    return
                field_name = get_fk_name_to_tracked_model(instance)
                log = instance._wicked_historian_changes_for_related_m2m_models[field_name]
                for pk, value in log.items():
                    instance_with_tracked_history = value['instance']
                    new_value = get_m2m_value(instance_with_tracked_history)
                    by_pk = lambda x: x['pk']
                    if sorted(log[instance_with_tracked_history.pk]['previous_value'], key=by_pk) == sorted(new_value, key=by_pk):
                        continue
                    instance_with_tracked_history._wicked_historian_m2m_changes['new_value'] = new_value
                    cls.create_m2m_history(instance_with_tracked_history)

            return pre_action, post_action

        @classmethod
        def reverse_fk_handlers_factory(cls, field_for_factory: ManyToOneRel) -> Tuple[Callable, Callable]:
            def get_current_value(instance):
                if instance is None:
                    return None, None
                tracked_history_instance = getattr(instance, field_for_factory.field.name)
                if tracked_history_instance is not None:
                    return tracked_history_instance.pk, list(getattr(tracked_history_instance, field_for_factory.get_accessor_name()).all())
                return None, None

            def set_value(instance, record, name):
                pk, value = record
                if pk is not None:
                    instance._previous_values.setdefault(pk, {})[name] = value

            def _fill_previous_values(instance, instance_from_db=None):
                instance_from_db = instance_from_db or instance
                set_value(instance, get_current_value(instance_from_db), 'old_value')

            def pre_action(sender: Type[Model], instance: Model, **kwargs):
                previous_instance_from_db = cls.get_model_or_none(instance)
                instance._previous_values = {}
                instance._previous_instance_from_db = previous_instance_from_db
                _fill_previous_values(instance)
                _fill_previous_values(instance, instance_from_db=previous_instance_from_db)

            def post_action(sender: Type[Model], instance: Model, **kwargs):
                set_value(instance, get_current_value(instance), 'new_value')
                set_value(instance, get_current_value(instance._previous_instance_from_db), 'new_value')
                for pk, data in instance._previous_values.items():
                    try:
                        cls.create_history(
                            field_for_factory.model.objects.get(pk=pk),
                            [(field_for_factory.get_accessor_name(), (data.get('old_value', []), data.get('new_value', [])))],
                        )
                    except field_for_factory.model.DoesNotExist:
                        # Model is already deleted so creating history is not necessary
                        continue

            return pre_action, post_action

        @classmethod
        def register_m2m_signals(cls):
            # noinspection PyProtectedMember
            # pylint: disable=protected-access
            for field in model_class._meta.get_fields():
                if field.many_to_many and not field.auto_created:
                    through_model = getattr(getattr(model_class, field.name), 'through')
                    m2m_changed.connect(
                        cls.m2m_changed,
                        sender=through_model,
                        dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                    )
                    if not through_model._meta.auto_created:
                        pre_function, post_function = cls.custom_through_m2m_handlers_factory(model_class, field)
                        pre_save.connect(
                            pre_function,
                            sender=through_model,
                            weak=False,
                            dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                        )
                        pre_delete.connect(
                            pre_function,
                            sender=through_model,
                            weak=False,
                            dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                        )
                        post_delete.connect(
                            post_function,
                            weak=False,
                            sender=through_model,
                            dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                        )
                        post_save.connect(
                            post_function,
                            sender=through_model,
                            weak=False,
                            dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                        )
                elif (
                    isinstance(field, ManyToOneRel) and
                    field.get_accessor_name() in getattr(model_class._meta, 'reverse_foreign_key_relations', set())
                ):
                    pre_function, post_function = cls.reverse_fk_handlers_factory(field)
                    pre_save.connect(
                        pre_function,
                        sender=field.related_model,
                        weak=False,
                        dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                    )
                    pre_delete.connect(
                        pre_function,
                        sender=field.related_model,
                        weak=False,
                        dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                    )
                    post_delete.connect(
                        post_function,
                        weak=False,
                        sender=field.related_model,
                        dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                    )
                    post_save.connect(
                        post_function,
                        sender=field.related_model,
                        weak=False,
                        dispatch_uid='{}__{}'.format(cls.__name__, field.name),
                    )

                    # Decorate reverse relation descriptor to prevent using manager methods with bulk=True
                    # This methods use update rather than save.

                    related_descriptor = getattr(model_class, field.get_accessor_name())

                    for method_name in ['add', 'clear', 'remove', 'set']:
                        method = getattr(related_descriptor.related_manager_cls, method_name, None)
                        if method is not None:
                            decorated_method = bulk_mode_check(method, field.get_accessor_name())
                            setattr(related_descriptor.related_manager_cls, method_name, decorated_method)



            cls.m2m_signals_registered = True

        def __str__(self) -> str:
            return u'{0} {1} - {2}: {3} => {4}'.format(
                self.user,
                self.change_date.strftime('%Y-%m-%d %H:%M:%S'),
                self.field_name,
                self.old_value,
                self.new_value,
            )

        @property
        def field_verbose_name(self) -> str:
            return field_choices_by_id[self.field].verbose_name

        @property
        def field_name(self) -> str:
            return field_choices_by_id[self.field].name

    attributes = {
        '__module__': module,
    }

    prefix = ''
    if abstract:
        class AbstractMeta(object):
            abstract = True
        attributes['Meta'] = AbstractMeta
        prefix = 'Base'

    return type('%s%sEditHistory' % (prefix, model_class.__name__), (EditHistory, ), attributes)


def method_singledispatch(func: Callable) -> Callable:
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)

    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


def handle_none_values(func: Callable) -> Callable:
    """Special decorator for wrapping method_singledispatch in DefaultFieldValueMapper."""

    def wrapper(self, field, value):
        if value is None:
            return JSON_NULL
        return func(self, field, value)

    update_wrapper(wrapper, func)
    return wrapper


class DefaultFieldValueMapper:

    @handle_none_values
    @method_singledispatch
    def __call__(self, field: models.Field, value: Any) -> ValueRepresentation:
        return value

    @__call__.register(models.ForeignKey)
    def get_foreign_key_representation(self, field: models.ForeignKey, pk_value: Any) -> ModelInstanceRepresentation:
        # noinspection PyProtectedMember
        # pylint: disable=protected-access
        related_object = field.related_model._default_manager.get(pk=pk_value)
        return self._get_model_instance_representation(related_object)

    @__call__.register(models.ManyToManyField)
    @__call__.register(ReverseForeignKeyRelation)
    def get_many_to_many_field_representation(
            self,
            field: models.ManyToManyField,   # pylint: disable=unused-argument
            instances: Iterable[models.Model]) -> Iterable[ModelInstanceRepresentation]:
        return [self._get_model_instance_representation(instance) for instance in sorted(instances, key=lambda instance: instance.pk)]

    @__call__.register(models.DateField)
    @__call__.register(models.DateTimeField)
    def get_date_representation(self, _: Union[models.DateField, models.DateTimeField], value: date) -> str:
        return value.isoformat()

    @__call__.register(models.DurationField)
    def get_timedelta_representation(self, _: models.DateField, value: timedelta) -> int:
        return int(value.total_seconds() * 1000)

    @__call__.register(models.FileField)
    def get_file_representation(self, _: models.FileField, value: FieldFile) -> str:
        try:
            result = value.url
        except ValueError:
            result = ''
        return result

    @__call__.register(models.BinaryField)
    def get_bytes_representation(self, _: models.BinaryField, value: bytes) -> str:
        return base64.b64encode(value).decode('utf-8')

    @__call__.register(models.DecimalField)
    def get_decimal_representation(self, _: models.DecimalField, value: Decimal) -> str:
        return str(value)

    @__call__.register(models.TimeField)
    def get_time_representation(self, _: models.TimeField, value: time):
        return value.isoformat()

    def _get_model_instance_representation(self, instance: models.Model) -> ModelInstanceRepresentation:
        # noinspection PyProtectedMember
        pk_field = instance._meta.pk  # pylint: disable=protected-access
        return {
            'pk': self(pk_field, instance.pk),
            'str': force_text(instance),
        }


def get_concrete_model_subclasses(model: Type[BaseModel]) -> Iterable[Type[BaseModel]]:
    # noinspection PyProtectedMember
    if not model._meta.abstract:  # pylint: disable=protected-access
        yield model

    for child_model in model.__subclasses__():
        for concrete_child_model in get_concrete_model_subclasses(child_model):
            yield concrete_child_model


def get_field_choices_from_model(model: Type[models.Model]) -> List[FieldDescription]:
    # noinspection PyProtectedMember
    result = []

    for field_instance in model._meta.fields + model._meta.many_to_many:  # pylint: disable=protected-access
        result.append(FieldDescription(field_instance.name, field_instance))

    return result


def bulk_mode_check(func, field_name):
    """Check if bulk kwarg is supplied in func call and have value False."""

    @wraps(func)
    def wraper(*args, **kwargs):
        assert kwargs.get('bulk', True) is False, (
            "You can't use related manager method \"{method_name}\" with bulk=True (which is default)"
            " when creating history for \"{field_name}\" field is enabled!"
        ).format(
            field_name=field_name,
            method_name=func.__name__,
        )
        return func(*args, **kwargs)

    return wraper



from django.apps import AppConfig
from django.db.models.signals import post_migrate


class WickedHistorianAppConfig(AppConfig):

    name = 'wicked_historian'

    def ready(self):
        super().ready()
        from . import checks  # noqa
        from wicked_historian.handlers import create_choices_db_view
        post_migrate.connect(create_choices_db_view, sender=self)



import typing
from typing import (
    Any,
    Dict,
    Iterable,
    List,
)

from diffable.models import DiffableModel
from django.db.models import (
    Model,
    fields,
    options,
)
from django.db.models.signals import pre_save
from django.utils.module_loading import import_string

from .deletion import DeletionGuard

if typing.TYPE_CHECKING:
    from .utils import (
        FieldDescription,
        ModelDiffItem,
    )


options.DEFAULT_NAMES = options.DEFAULT_NAMES + ('history_class', 'reverse_foreign_key_relations',)


class DiffableHistoryModel(DiffableModel):

    _wicked_historian_diff_items = None
    deletion_guard = DeletionGuard()

    class Meta:
        abstract = True
        history_class = None

    def save(self, *args, **kwargs):
        if not self._meta.history_class:
            raise NotImplementedError('Specify `history_class` meta property')
        history_class = import_string(self._meta.history_class)
        pre_save_dispatch_uid = 'store_diff_items_{}'.format(id(self))
        pre_save.connect(self.store_diff_items, sender=self.__class__, dispatch_uid=pre_save_dispatch_uid)
        try:
            super(DiffableHistoryModel, self).save(*args, **kwargs)
        finally:
            pre_save.disconnect(dispatch_uid=pre_save_dispatch_uid)
        history_class.create_history(self, diff_items=self._wicked_historian_diff_items)

    def store_diff_items(self, sender, instance, **kwargs):
        """A pre_save signal handler that is going to be attached as the last pre_save handler - stores the diff in the instance."""
        if instance is not self:
            return
        diff_items = list(self.diff.items())
        if not self.pk:
            for field in self._meta.fields:
                if (not self._state.adding and
                        field.name not in self.changed_fields and field.default != fields.NOT_PROVIDED and field.default is not None):
                    diff_items.append(
                        (field.name, (None, field.default))
                    )
        self._wicked_historian_diff_items = diff_items

    def delete(self, using=None, keep_parents=False):
        with self.deletion_guard.deletion_context(self):
            return super(DiffableHistoryModel, self).delete(using=using, keep_parents=keep_parents)


class BaseEditHistory(Model):
    """Base abstract history model for diffable model edits."""

    class Meta:
        abstract = True

    m2m_signals_registered = False

    class UnknownFieldException(Exception):
        pass

    class FieldNotTracked(KeyError):
        pass

    @classmethod
    def get_for(cls, instance: Model) -> Iterable[Dict[str, Any]]:
        raise NotImplementedError()

    @classmethod
    def get_history_entry(cls, entry: 'BaseEditHistory') -> Dict[str, Any]:
        raise NotImplementedError()

    @classmethod
    def get_tracked_field_choice_by_name(cls, name: str) -> 'FieldDescription':
        raise NotImplementedError()

    @classmethod
    def get_value_representation(cls, name: str, value: Any) -> Any:
        raise NotImplementedError()

    @classmethod
    def create_history(cls, instance: Model, diff_items: List['ModelDiffItem']):
        raise NotImplementedError()

    @classmethod
    def create_m2m_history(cls, instance: Model):
        raise NotImplementedError()

    @classmethod
    def m2m_changed(cls, sender, **kwargs):
        raise NotImplementedError()

    @classmethod
    def register_m2m_signals(cls):
        raise NotImplementedError()

    def __str__(self) -> str:
        raise NotImplementedError()

    @property
    def field_verbose_name(self) -> str:
        raise NotImplementedError()

    @property
    def field_name(self) -> str:
        raise NotImplementedError()



from typing import (
    Any,
    Dict,
    Iterable,
    Set,
    Tuple,
    Type,
)

from django.apps import apps
from django.core.checks import (
    CheckMessage,
    Error,
    Tags,
    Warning,
    register,
)
from django.conf import settings
from django.db import (
    connection,
    models,
)
from django.db.migrations import operations
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.graph import MigrationGraph
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.operations.models import ModelOperation
from django.db.migrations.questioner import NonInteractiveMigrationQuestioner
from django.db.migrations.state import ProjectState

from .models import BaseEditHistory
from .utils import get_concrete_model_subclasses


Choice = Tuple[Any, str]

check_missing_migration = getattr(settings, 'WICKED_HISTORIAN_RUN_CHECK_MISSING_MIGRATION', True)


@register(Tags.signals)
def check_m2m_registration_for_edit_history(**kwargs) -> Iterable[CheckMessage]:
    messages = []
    for history_model in get_concrete_model_subclasses(BaseEditHistory):
        if not history_model.m2m_signals_registered:
            messages.append(
                Error(
                    "M2M signals are not registered for {} model!".format(history_model.__name__),
                    hint="Add {}.register_m2m_signals call to your AppConfig.ready.".format(history_model.__name__),
                    obj=history_model,
                    id='wicked_historian.E001',
                )
            )
    return messages


@register
def check_incremental_changes_in_field_choices(**kwargs) -> Iterable[CheckMessage]:
    messages = []
    migration_loader = MigrationLoader(None, ignore_no_migrations=True)
    for history_model in get_concrete_model_subclasses(BaseEditHistory):
        # noinspection PyProtectedMember
        # pylint: disable=protected-access
        app_label = history_model._meta.app_label
        if is_any_migration_missing(migration_loader, app_label):
            messages.append(
                Warning(
                    'Cannot check incremental changes in fields - there are changes in the {} app without a migration.'.format(app_label),
                    hint='Run makemigrations management command.',
                    obj=history_model,
                    id='wicked_historian.W001',
                )
            )
        else:
            removed_fields = get_removed_choices(migration_loader.graph, history_model, 'field')
            if removed_fields:
                messages.append(
                    Error(
                        'Detected a removed field: {}'.format(removed_fields),
                        hint='Move the old versions of fields to the `obsolete_field_choices` list.',
                        obj=history_model,
                        id='wicked_historian.E002',
                    )
                )
    return messages


@register
def check_incremental_changes_in_choices_of_fields(**kwargs) -> Iterable[CheckMessage]:
    messages = []
    migration_loader = MigrationLoader(None, ignore_no_migrations=True)
    for history_model in get_concrete_model_subclasses(BaseEditHistory):
        # noinspection PyProtectedMember
        # pylint: disable=protected-access
        history_model_opts = history_model._meta
        app_label = history_model_opts.app_label
        if is_any_migration_missing(migration_loader, app_label):
            messages.append(
                Warning(
                    'Cannot check incremental changes in choices of fields - there are changes in the {} app without a migration.'.format(app_label),
                    hint='Run makemigrations management command.',
                    obj=history_model,
                    id='wicked_historian.W002',
                )
            )
        else:
            tracked_model = history_model_opts.get_field('model').related_model
            # noinspection PyProtectedMember
            # pylint: disable=protected-access
            tracked_model_opts = tracked_model._meta
            for field in tracked_model_opts.fields:
                removed_choices = get_removed_choices(migration_loader.graph, tracked_model, field.name)
                if removed_choices:
                    messages.append(
                        Error(
                            'Detected removed choices of field {}: {}'.format(field.name, removed_choices),
                            hint='Models with history tracking need to have incremental changes in their field choices.',
                            obj=tracked_model,
                            id='wicked_historian.E003',
                        )
                    )
    return messages


def get_removed_choices(migration_graph: MigrationGraph, model: Type[models.Model], field_name: str) -> Set[Choice]:
    # noinspection PyProtectedMember
    # pylint: disable=protected-access
    current_choices = get_current_choices(model, field_name)
    historical_choices = get_all_historical_choices(migration_graph, model, field_name)

    current_choices_values = set(choice[0] for choice in current_choices)
    historical_choices_values = set(choice[0] for choice in historical_choices)
    return historical_choices_values - current_choices_values


def get_current_choices(model: Type[models.Model], field_name: str) -> Set[Choice]:
    model_opts = model._meta
    return set(model_opts.get_field(field_name).choices)


def get_all_historical_choices(migration_graph: MigrationGraph, model: Type[models.Model], field_name: str) -> Set[Choice]:
    # noinspection PyProtectedMember
    # pylint: disable=protected-access
    options = model._meta
    assert field_name in (field.name for field in options.fields), 'Field not found in model fields.'

    app_label = options.app_label
    models_fields_all_choices = get_all_fields_all_historical_choices(migration_graph, app_label)
    # We are iterating though all current fields and since checks are run before making migrations we need to get(field_name, [])
    # so that we don't crash when adding new field without historical choices
    return models_fields_all_choices[model.__name__.lower()].get(field_name, [])


def update_choices_by_model_name_by_field_name(choices_by_model_name_by_field_name: Dict[str, Dict[str, Set[Choice]]], operation: ModelOperation):
    def get_field_choices(model_field: models.Field) -> Set[Choice]:
        return set(model_field.choices or [])

    if isinstance(operation, operations.CreateModel):
        choices_by_model_name_by_field_name[operation.name.lower()] = {}
        for operation_field_name, field in operation.fields:
            choices_by_model_name_by_field_name[operation.name.lower()][operation_field_name] = get_field_choices(field)
    elif isinstance(operation, operations.AddField):
        choices_by_model_name_by_field_name[operation.model_name][operation.name] = get_field_choices(operation.field)
    elif isinstance(operation, operations.AlterField):
        choices_by_model_name_by_field_name[operation.model_name][operation.name] |= get_field_choices(operation.field)
    elif isinstance(operation, operations.RenameField):
        choices_by_model_name_by_field_name[operation.model_name][operation.new_name] = (
            choices_by_model_name_by_field_name[operation.model_name][operation.old_name]
        )
        del choices_by_model_name_by_field_name[operation.model_name][operation.old_name]
    elif isinstance(operation, operations.RenameModel):
        choices_by_model_name_by_field_name[operation.new_name.lower()] = choices_by_model_name_by_field_name[operation.old_name.lower()]
        del choices_by_model_name_by_field_name[operation.old_name.lower()]
    elif isinstance(operation, operations.RemoveField):
        del choices_by_model_name_by_field_name[operation.model_name][operation.name]
    elif isinstance(operation, operations.DeleteModel):
        del choices_by_model_name_by_field_name[operation.name.lower()]


def get_all_fields_all_historical_choices(migration_graph: MigrationGraph, app_label: str) -> Dict[str, Dict[str, Set[Choice]]]:
    choices_by_model_name_by_field_name = {}

    related_app_labels = {k[0] for k, v in migration_graph.nodes.items() if v.app_label == app_label}
    dependencies = [leaf_node for app_label in related_app_labels for leaf_node in migration_graph.leaf_nodes(app_label)]

    executor = MigrationExecutor(connection)
    plan = executor.migration_plan(dependencies, clean_start=True)

    for migration, _ in plan:
        if migration.app_label not in related_app_labels:
            continue
        for operation in migration.operations:
            update_choices_by_model_name_by_field_name(choices_by_model_name_by_field_name, operation)

    return choices_by_model_name_by_field_name


class ExceptionRaisingNonInteractiveMigrationQuestioner(NonInteractiveMigrationQuestioner):

    """Custom NonInteractiveMigrationQuestioner that raises an exception instead of calling `sys.exit`."""

    class InteractionNeeded(Exception):
        pass

    def ask_not_null_addition(self, field_name, model_name):
        raise self.InteractionNeeded()

    def ask_auto_now_add_addition(self, field_name, model_name):
        raise self.InteractionNeeded()


def is_any_migration_missing(migration_loader: MigrationLoader, app_label: str) -> bool:
    if not check_missing_migration:
        return False

    if app_label not in is_any_migration_missing.results_cache:
        questioner = ExceptionRaisingNonInteractiveMigrationQuestioner(specified_apps=[app_label], dry_run=True)
        autodetector = MigrationAutodetector(
            migration_loader.project_state(),
            ProjectState.from_apps(apps),
            questioner,
        )
        try:
            changes = autodetector.changes(
                graph=migration_loader.graph,
                trim_to_apps=[app_label],
                convert_apps=[app_label],
                migration_name='check',
            )
        except ExceptionRaisingNonInteractiveMigrationQuestioner.InteractionNeeded:
            is_any_migration_missing.results_cache[app_label] = True
        else:
            is_any_migration_missing.results_cache[app_label] = bool(changes)
    return is_any_migration_missing.results_cache[app_label]


is_any_migration_missing.results_cache = {}  # type: Dict[str, bool]



import threading
from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
    Type,
)
from contextlib import contextmanager

if TYPE_CHECKING:
    from django.db.models import Model


class DeletionGuard(threading.local):

    def __init__(self):
        super().__init__()
        self.models_during_deletion = set()

    def start_deletion(self, identifier: Tuple[Type['Model'], Any]):
        self.models_during_deletion.add(identifier)

    def end_deletion(self, identifier: Tuple[Type['Model'], Any]):
        self.models_during_deletion.remove(identifier)

    def create_identifier(self, instance: 'Model') -> Tuple[Type['Model'], Any]:
        return instance.__class__, instance.pk

    def is_instance_during_deletion(self, instance: 'Model') -> bool:
        return self.create_identifier(instance) in self.models_during_deletion

    @contextmanager
    def deletion_context(self, instance: 'Model') -> 'DeletionGuard':
        identifier = self.create_identifier(instance)
        try:
            self.start_deletion(identifier)
            yield self
        finally:
            self.end_deletion(identifier)



# -*- coding: utf-8 -*-
import os
from setuptools import find_packages, setup


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-wicked-historian',
    version='1.0.0',
    description='A complete solution for creating automatic history of Django models.',
    long_description=README,
    long_description_content_type='text/markdown',
    author='IIIT',
    author_email='github@iiit.pl',
    packages=find_packages(exclude=[
        'testproject',
        'testproject.*',
    ]),
    include_package_data=True,
    install_requires=[
        'Django>=2.1,<3.0',
        'django-diffable>=1.0.0'
    ],
    extras_require={
        'mysql': ['django-mysql>=2.2.0'],
        'postgres': ['psycopg2>=2.6.2'],
        'django-jsonfield': ['django-jsonfield>=1.0.1'],
    },
    tests_require=[
        'freezegun>=0.3.10',
        'django-jsonfield>=1.0.1',
        'pytz>=2018.3',
        'factory_boy==2.10.0',
        'django-override-storage>=0.1.2',
    ],
    test_suite='testproject.runtests.run_tests',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.10',
    ],
)



from django.db.models.lookups import (
    Contains,
    Exact,
    GreaterThan,
    GreaterThanOrEqual,
    IExact,
    LessThan,
    LessThanOrEqual,
    Lookup,
)

from django_mysql.models.fields.json import (
    KeyTransform,
    KeyTransformFactory,
)

from .mysql import JSONField as MySQLJSONField

__all__ = (
    'JSONField',
)


class JSONField(MySQLJSONField):
    
    """MySQL JSONField with lookups fixed to work with MariaDB.

    Using MySQLJSONField won't work on MariaDB due to MySQLJSONField creating sql queries with "(CAST(... AS JSON)))", which MariaDB rejects
    as invalid syntax. In order to work around this issue, this JSONField reregisters django_mysql lookups with custom JSONLookupMixin.

    Furthermore, this class supports searching through json arrays with `arrayelements` transform. See KeyTransformWithArrayMemberLookupSupport for
    more information.
    """

    def get_transform(self, name):
        transform = super(JSONField, self).get_transform(name)
        if isinstance(transform, KeyTransformFactory):
            return MyKeyTransformFactory(name)
        if transform:
            return transform  # pragma: no cover

        return MyKeyTransformFactory(name)


class JSONLookupMixin(object):
    
    def get_prep_lookup(self):
        value = self.rhs
        if not hasattr(value, '_prepare') and value is not None and isinstance(value, str):
            return '"{}"'.format(value)
        return super().get_prep_lookup()


@JSONField.register_lookup
class JSONFieldExact(JSONLookupMixin, Exact):
    pass


@JSONField.register_lookup
class JSONFieldIExact(JSONLookupMixin, IExact):
    pass


@JSONField.register_lookup
class JSONFieldContains(JSONLookupMixin, Lookup):
    """Modified version of django_mysql.models.lookups.JSONContains that does not use JSONLookupMixin."""

    lookup_name = 'contains'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return 'JSON_CONTAINS({}, {})'.format(lhs, rhs), params


@JSONField.register_lookup
class JSONFieldGreaterThan(JSONLookupMixin, GreaterThan):
    pass


@JSONField.register_lookup
class JSONFieldGreaterThanOrEqual(JSONLookupMixin, GreaterThanOrEqual):
    pass


@JSONField.register_lookup
class JSONFieldLessThan(JSONLookupMixin, LessThan):
    pass


@JSONField.register_lookup
class JSONFieldLessThanOrEqual(JSONLookupMixin, LessThanOrEqual):
    pass


class KeyTransformWithArrayMemberLookupSupport(KeyTransform):
    """Custom subclass django_mysql's KeyTransform with support for searching through JSON array elements using wildcard in jsonpath.

    Using arrayelements transform is necessary for searching through JSONFields with arrays inside. You have to use `contains` lookup at the end
    in order to check if any of the elements matches certain value.
    Example:
    >>> ModelCls.objects.create(example_json_field=[{'pk': 1}, {'pk': 2}])
    >>> assert ModelCls.objects.filter(example_json_field__arrayelements__pk__contains=2).count() == 1
    >>> assert ModelCls.objects.filter(example_json_field__arrayelements__pk__contains=3).count() == 0

    """

    list_contains_lookup_name = 'arrayelements'

    def compile_json_path(self, key_transforms):
        path = ['$']
        for key_transform in key_transforms:
            try:
                if key_transform == self.list_contains_lookup_name:
                    path.append('[*]')
                else:
                    num = int(key_transform)
                    path.append('[{}]'.format(num))
            except ValueError:  # non-integer
                path.append('.')
                path.append(key_transform)
        return ''.join(path)


class MyKeyTransformFactory(object):

    def __init__(self, key_name):
        self.key_name = key_name

    def __call__(self, *args, **kwargs):
        return KeyTransformWithArrayMemberLookupSupport(self.key_name, *args, **kwargs)
